#!/bin/bash

shellfu import pretty

#
# TaskWarrior smart wrapper
#
# This library provides internal functions for `bmo-qtask`, which is a smart
# wrapper around TaskWarrior command, providing you with an extensible way
# to manage creation of sets of tasks.
#
#
# Motivation
# ==========
#
#
# For example, imagine you have a bug trakser named Foo.  Each time you are
# assigned a task, this means several sub-tasks:
#
#  *  triage
#  *  prepare fixed build
#  *  test
#  *  submit build
#  *  switch to QE.
#
# Obviously, turning all of the above can quickly turn into chore, especially so
# if fixing bugs is your daily bread.
#
# One way you can avoid the chores is write a custom script, but that's work that
# you would have to do, and worse, does not account for cases when there are
# *more* types of items you would like to handle:  customer cases, research topics,
# courses?
#
# bmo-qtask allows you to write a plugin for each of the task type you normally
# encounter and then use qtask's universal interface for *any* type of task.
# These plugins can then be easily shared with colleagues, extending possibilities
# to save work.  In particular, it enables you to say:
#
#     bmo qtask "http://foo.internal/bug/12345"
#
# and effectively create task tree for your workflow:
#
#     bug 12345
#     └── switch to QE
#         └── submit build
#             └── test
#                 └── prepare build
#                     └── triage
#
# Moreover, if your Foo tracker (or other interconnected systems) provides
# a way to access various meta-data such as project names, contacts or due
# dates, your plugin can further append that to the relevant tasks in this
# sub-tree, making it even more useful and reducing need to switch back/forth
# between many systems just to jump-start your actual work on tasks.
#
#
# Task Tree Definition Language
# =============================
#
# The interface between this library and the plugin is based on "Task
# Tree Definition Language" (TTDL).  The syntax is very simple, consisting
# of plain text lines, one line for one property definition:
#
#     [%N:]NAME:[VALUE]
#
# where:
#
#  *  N is a subtask ID -- an arbitrary unique integer,
#  *  NAME is a property name,
#  *  and VALUE is any value used for that property.
#
#
# Properties
# ----------
#
# Following property names are defined:
#
#  *  `ANNOTATION` - task annotation,
#  *  `DUE`        - due date,
#  *  `PROJECT`    - project name,
#  *  `TAG`        - tag to add,
#  *  `TITLE`      - task title (mandatory)
#
#
# Defining task
# -------------
#
# The minimum definition looks like this:
#
#     TITLE:some task
#
# Attributes such as due dates are added as following:
#
#     TITLE:some task
#     DUE:2017-05-05
#
# Sub-tasks are defined using the N prefix.  Each sub-task must have
# a hexadecimal id unique within a definition.  What follows the prefix
# has exactly the same syntax as above.  For example:
#
#     TITLE:some task
#     DUE:2017-05-05
#     %1:TITLE:subtask
#     %1:TAG:foo
#     %2:TITLE:another subtask
#     %2:TAG:foo
#
# Above, we have added two sub-tasks, "subtask" and "another subtask".
# This notation will result in 3 tasks, where the "main" task "some task"
# depends on two tasks: "subtask" and "another subtask".
#
# Notice two things:
#
#  *  The values "1" and "2" are insignificant, we could have used any
#     other numbers as long as they are not the same (it's perfectly fine
#     to use random numbers here).
#
#  *  Since what follows the prefix has identical syntax as the whole line,
#     it can also have a prefix!  This way we can recursively define trees
#     of tasks:
#
#         TITLE:some task
#         DUE:2017-05-05
#         %1:TITLE:subtask
#         %1:TAG:foo
#         %1:%3:TITLE:subtask's subtask
#         %2:TITLE:another subtask
#         %2:TAG:foo
#
#     The above represents tree of tasks:
#
#         some task
#         ├── subtask
#         │   └── subtask's subtask
#         └── another subtask
#


#
# TaskWarrior command
#
BMO_QTASK_TASKCMD=${BMO_QTASK_TASKCMD:-task}


_bmo_qtask__get_ttdl() {
    #
    # Ask plugin $Plugin for task definition
    #
    # Function ${Plugin}__define is called and must respond with definition
    # Task Tree Definition Language.  See respective section of this library's
    # manual.
    #
    local deffn     # name of plugin's definition function
    local deffn="${Plugin}__define"
    type -t "$deffn" >/dev/null || {
        warn "plugin define function missing: $deffn"
        return 3
    }
    "$deffn"
}

_bmo_qtask__mktree() {
    #
    # Make tree of tasks
    #
    local tmp       # temporary directory
    tmp=$(mktemp -d -t bmo_qtask__mktree.XXXXXXXX)
    pushd "$tmp" >/dev/null || die
        _bmo_qtask__mktree_do
    popd >/dev/null || die
    rm -r "$tmp"
}

_bmo_qtask__mktree_do() {
    #
    # Really make tree of tasks
    #
    local parent_uuid   # parent task's UUID
    local child_id      # child's local id
    cat > tree.ttdl
    <tree.ttdl grep '^[A-Z]' > parent.ttdl
    test -s parent.ttdl || {
        warn "nothing in this branch, skipping"
        return 3
    }
    parent_uuid=$(_bmo_qtask__new <parent.ttdl)
    debug -v parent_uuid
    grep -o '^%[0-9A-Fa-f]\+' tree.ttdl \
      | sort \
      | uniq \
      > children
    while read -r child_id
    do
        debug -v child_id
        grep "^$child_id:" tree.ttdl \
          | cut -d: -f2- \
          > child.ttdl
        <child.ttdl _bmo_qtask__mktree
    done <children > children.uuids
    for subuuid in $(<children.uuids);
    do
        echo "DEPENDS:$subuuid" | _bmo_qtask__setprops "$parent_uuid"
    done
    echo "$parent_uuid"
    cat children.uuids
}

_bmo_qtask__new() {
    #
    # Create new tasks from definition on STDIN, echo its UUID then
    #
    local uuid          # new task's UUID
    uuid=$(_bmo_qtask__new_do)
    debug -v uuid
    _bmo_qtask__setprops "$uuid"
    echo "$uuid"
}

_bmo_qtask__new_do() {
    #
    # Add a new dummy task and print its UUID
    #
    local id        # our dummy id
    id="bmo_qtask_dummy_$(uuidgen | cut -d- -f1)"
    debug -v id
    "$BMO_QTASK_TASKCMD" add -- "bmo_qtask.id=$id" >/dev/null
    #
    # To avoid matching thrash, we create (using 'rc.*' arguments) ad-hoc
    # report type BMO_QTASK_UUID.  In TaskWarrior, the report type is
    # used directly as command, hence the last line.
    #
    "$BMO_QTASK_TASKCMD" rc.context:none rc.verbose:none \
        rc.report.BMO_QTASK_UUID.columns:uuid \
        rc.report.BMO_QTASK_UUID.labels:BMO_QTASK_UUID \
        description.has:"bmo_qtask.id=$id" \
        BMO_QTASK_UUID \
        2>/dev/null \
      | grep -e '-[0-9a-f][0-9a-f]*$'
}

_bmo_qtask__setprops() {
    #
    # Set all properties on stdin to task UUID $1
    #
    local Uuid=$1                       # task task's UUID
    local line                          # a line of definition
    local name                          # name of property
    local value                         # value of property
    local do=_bmo_qtask__setprops_do    # worker command alias
    local q="printf %q"                 # quoting command alias
    while read -r line
    do
        name=${line%%:*}
        value=${line#"$name:"}
        case $name in
            ANNOTATION) $do annotate    "$value"                    ;;
            DEPENDS)    $do modify      depends:"$($q "$value")"    ;;
            DUE)        $do modify      due:"$($q "$value")"        ;;
            PROJECT)    $do modify      project:"$($q "$value")"    ;;
            TAG)        $do modify      "+$value"                   ;;
            TITLE)      $do modify      "$value"                    ;;
            *) warn "ignoring unknown property: $name" ;;
        esac
    done
}

_bmo_qtask__setprops_do() {
    #
    # Do $@ silently with task $Uuid
    #
    "$BMO_QTASK_TASKCMD" "$Uuid" "$@" >&/dev/null
}

bmo_qtask__choose_plugin() {
    #
    # Choose which plugin to use for data $BMO_QTASK_DATA
    #
    # Load each possible plugin (ask bmo_qtask__find_plugins()) and call
    # function named bmo_qtask_NAME__check() with the data fed to standard
    # input.  Exit status of zero means that the plugin supports this
    # data and is chosen.
    #
    # Name of chosen plugin is printed to stdout.
    #
    test -n "$BMO_QTASK_DATA" || {
        warn "BMO_QTASK_DATA is empty"
        return 3
    }
    local plname
    local compatfn
    for plname in $(bmo_qtask__find_plugins);
    do
        debug -v plname
        shellfu import "${plname}" || {
            warn "failed to import plugin: $plname"
            continue
        }
        compatfn="${plname}__check"
        type -t "$compatfn" >/dev/null || {
            warn "missing check function in plugin: $plname"
            continue
        }
        "$compatfn" <<<"$BMO_QTASK_DATA" >/dev/null || continue
        echo "$plname"
        return 0
    done
    return 1
}

bmo_qtask__define() {
    #
    # Print whole TTDL
    #
    local Plugin=$1             # plugin module of choice
    debug -v Plugin
    shellfu import "$Plugin"
    _bmo_qtask__get_ttdl
}

bmo_qtask__exec() {
    #
    # Execute TTDL on standard input
    #
    local uuid                  # each UUID of created task
    _bmo_qtask__mktree \
      | while read -r uuid;
        do
            "$BMO_QTASK_TASKCMD" "$uuid" info
        done
}

bmo_qtask__find_plugins() {
    #
    # List all installed plugins
    #
    shellfu _list_mfiles \
      | sed 's|\.sh$||; s|.*/||;' \
      | grep -x 'bmo_qtask_.*'
    #FIXME: hack: _list_mfiles is not a public API
}
